/* libfoXdesktop - foXdesktop's core library extending FOX
 * Copyright (C) 2000-2001 Josua Groeger <fractalus@gmx.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "FDComponentConstructDict.h"
#include "fd_defaults.h"
#include <iostream.h>

FXIMPLEMENT (FDComponentConstructDict, FXDict, NULL, 0);

FDComponentConstructDict::FDComponentConstructDict (FDApp *a): FXDict ()
{
  app= a;
  lt_dlinit ();
};

FDComponentConstructDict::~FDComponentConstructDict ()
{
  lt_dlexit ();
};

void *FDComponentConstructDict::createData (const void *ptr)
{
  lt_dlhandle *hdl= new lt_dlhandle;
  *hdl= *((lt_dlhandle*) ptr);
  return (void*) hdl;
}
    
void FDComponentConstructDict::deleteData (void *ptr)
{
  lt_dlclose (*((lt_dlhandle*) ptr));
  delete (lt_dlhandle*) ptr;
};

/* insert -----------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------------*/

TFDCConstruct *FDComponentConstructDict::insert (const FXchar *type, const FXchar *component, FXbool mrk= false)
{ 
  // Calculate the component library's file name
  FXString fdcpath (app->reg ().readStringEntry ("SETTINGS", "fdcpath", fd_default_fdcpath));
  FXString path, name;
  for (unsigned short i= 0; (path= fdcpath.extract (i, ':'))!= ""; i++)
  {
    name= path;
    name+= FXString ("/")+ FXString (type)+ FXString ("/libFDC")+ FXString (component)+ FXString (".so");
    if (FXFile::exists (name)) break; else name= "";
  }
  
  // Open the file containing the component
  TFDCConstruct *FDCConstruct= NULL;
  if (name== "") cerr<< "The foXdesktop component "<< type<< "/"<< component<< " could not be found!\n";
  else
  {
    // Get the handle
    lt_dlhandle hndl= lt_dlopen (name.text ());
    if (!hndl) cerr<< lt_dlerror ()<< "\n";
    else
    {
      // Get the component's interface version
      name= "FDC";
      name+= component;
      name.append ("_interface");
      const int* interface= (const int*) lt_dlsym (hndl, name.text ());
      
      // Compare the interfaces
      if (!interface) cerr<< "The foXdesktop component "<< type<< "/"<< component<< " provides no versioning information!\n";
      
      else if (*interface> FDC_INTERFACE) // Interface too new
        cerr<< "The foXdesktop component "<< type<< "/"<< component<<
               "\nhas an interface that is too new!\nThe interface version must be exactly "<<
               FDC_INTERFACE_MAJOR<< "."<< FDC_INTERFACE_MINOR<< "."<< FDC_INTERFACE_SUBMIN<<
               "!\nYou are recommended to upgrade libfoXdesktop.\n";
        
	else if (*interface< FDC_INTERFACE) // Interface too old
	  cerr<< "The foXdesktop component "<< type<< "/"<< component<<
                 "\nhas an interface that is too old!\nThe interface version must be exactly "<<
                 FDC_INTERFACE_MAJOR<< "."<< FDC_INTERFACE_MINOR<< "."<< FDC_INTERFACE_SUBMIN<<
                 "!\nYou are recommended to upgrade the component.\n";
        else
        {
          // Get the component's construct function
          name= "FDC";
	  name+= component;
          name.append ("_construct");
          FDCConstruct= (TFDCConstruct*) lt_dlsym (hndl, name.text ());
        
          // Add to dict
          if (!FDCConstruct) cerr<< "The foXdesktop component "<< type<< "/"<< component<< " provides no construction function!\n";
	  else
          {
            name= FXString (type)+ FXString ("/")+ FXString (component);
            FXDict::insert (name.text (), &hndl, mrk);
          }
        }
    }
  }
  return FDCConstruct;
}

TFDCConstruct *FDComponentConstructDict::insert (const FXchar *type, FXbool mrk= false)
{
  // Set lsStr e.g. to "ls -1 /usr/local/lib/foXdesktop-components/foXcontrol/libFDC*.so /usr/lib/foXdesktop-components/foXcontrol/libFDC*.so "
  // (if fdcpath is "/usr/local/lib/foXdesktop-components:/usr/lib/foXdesktop-components")
  FXString lsStr ("ls -1 ");
  lsStr.append (app->reg ().readStringEntry ("SETTINGS", "fdcpath", fd_default_fdcpath));
  lsStr+= ":";
  for (int i= 0; lsStr[i]; i++) if (lsStr[i]== ':') lsStr.replace (i, 1, "/foXcontrol/libFDC*.so ");
  
  // Try to open lsStr (find component libraries)
  TFDCConstruct *returnConstruct= NULL, *tmpConstruct= NULL;
  FILE *file= popen (lsStr.text (), "r"); 
  if (!file) perror ("popen");
  else
  {
    char fileName[1024];
    FXString componentName;
    // Parse the result of ls
    while (fgets (fileName, 1024, file))
    {
      // Cut spaces...
      char *ws= strpbrk(fileName, " \t\n");
      if (ws) *ws= 0;
      
      // Calculate the component name (from the file name)  
      componentName= fileName;
      componentName= componentName.after ('/', -1); 
      componentName.remove (0, 6); // remove "libFDC"
      componentName= componentName.before ('.', 1); // remove ".so"
      
      // Insert one component
      tmpConstruct= insert (type, componentName.text (), mrk);
      if (tmpConstruct) returnConstruct= tmpConstruct;
    }
  }    
  return returnConstruct; // Return the last valid FDCConstruct (respectively NULL)
}

/* remove, find, data -----------------------------------------------------------------------------
-------------------------------------------------------------------------------------------------*/

TFDCConstruct *FDComponentConstructDict::remove (const FXchar *type, const FXchar *component)
{
  FXString key (FXString (type)+ "/"+ FXString (component));
  return (TFDCConstruct*) FXDict::remove (key.text ());
}

TFDCConstruct *FDComponentConstructDict::find (const FXchar *type, const FXchar *component)
{
  FXString key (FXString (type)+ "/"+ FXString (component));
  FXString constructName ("FDC"+ FXString (component)+ "_construct");

  if (!FXDict::find (key.text ())) return NULL;
  else return (TFDCConstruct*) lt_dlsym (*((lt_dlhandle*) FXDict::find (key.text ())), constructName.text ());
}

TFDCConstruct *FDComponentConstructDict::data (FXuint pos)
{
  FXString constructName ("FDC"+ FXString (key (pos)).after ('/', -1)+ "_construct");
  return (TFDCConstruct*) lt_dlsym (*((lt_dlhandle*) FXDict::data (pos)), constructName.text ());
}
